import datetime
import logging
import time
import typing
from collections import namedtuple
from copy import copy
from enum import Enum

from construct import Container

from paradox.config import config as cfg
from paradox.data.enums import EventLevel
from paradox.lib.format import EventMessageFormatter

logger = logging.getLogger("PAI").getChild(__name__)


class Change:
    def __init__(
        self,
        type: str,
        key: str,
        property: str,
        new_value,
        old_value=None,
        initial=None,
    ):
        self.type = type
        self.key = key
        self.property = property
        self.new_value = new_value
        self.old_value = old_value
        self.initial = initial
        if self.initial is None:
            self.initial = old_value is None
        self.timestamp = int(time.time())

    def __eq__(self, other: "Change"):
        return (
            self.type == other.type
            and self.key == other.key
            and self.property == other.property
            and self.new_value == other.new_value
        )

    def __repr__(self):
        return "Change {}/{}/{} from {}{} to {}".format(
            self.type,
            self.key,
            self.property,
            self.old_value,
            ("(initial)" if self.initial else ""),
            self.new_value,
        )


class Event:
    def __init__(self, label_provider=None):
        self.timestamp = 0
        self.level = EventLevel.NOTSET
        self.type = None
        self._message_tpl = ""
        self.change = {}
        self.tags = []
        self.additional_data = {}
        self.hook_fn = None

        self._key = None

        if label_provider is not None:
            self.label_provider = label_provider
        else:
            self.label_provider = lambda type, value: "[{}:{}]".format(type, value)

    def __repr__(self):
        lvars = {}
        lvars.update(self.__dict__)
        lvars["message"] = self.message

        return (
            str(self.__class__)
            + "\n"
            + "\n".join(
                (
                    "{} = {}".format(item, lvars[item])
                    for item in lvars
                    if not item.startswith("_")
                )
            )
        )

    @property
    def key(self) -> str:
        if not self._key:
            self._key = "{},{},{}".format(
                self.type,
                self.label,
                ",".join("=".join([key, str(val)]) for key, val in self.change.items()),
            )

        return self._key

    @property
    def message(self) -> str:
        return EventMessageFormatter().format(self._message_tpl, self)

    @property
    def props(self) -> dict:
        dp = {}
        for key in dir(self):
            if key in ["props", "raw", "hook_fn"]:
                continue

            value = getattr(self, key)
            if not key.startswith("_") and not isinstance(value, typing.Callable):
                if isinstance(value, datetime.datetime):
                    dp[key] = str(value.isoformat())
                elif isinstance(value, Enum):
                    dp[key] = value.name
                else:
                    dp[key] = value
        return dp

    def call_hook(self, *args, **kwargs):
        if isinstance(self.hook_fn, typing.Callable):
            kwargs["event"] = self
            try:
                self.hook_fn(*args, **kwargs)
            except:
                logger.exception("Failed to call event hook")


class LiveEvent(Event):
    def __init__(self, event: Container, event_map: dict, label_provider=None):
        raw = event.fields.value
        if raw.po.command != 0x0E:
            raise AssertionError("Message is not an event")

        # parse event map
        if raw.event.major not in event_map:
            raise AssertionError("Unknown event major: {}".format(raw))

        super(LiveEvent, self).__init__(label_provider=label_provider)

        self.major = raw.event.major  # Event major code
        self.minor = raw.event.minor  # Event minor code
        if hasattr(raw.event, "minor2"):  # For EVO panels
            self.minor2 = raw.event.minor2

        # default
        self.id = None  # Element ID (if available in the event)
        self.additional_data = {}

        self.timestamp = raw.time  # Event timestamp
        self.partition = (
            raw.partition
        )  # Event Partition. if Type is system, partition may not be relevant
        self.module = raw.module_serial  # Event Module Serial
        self.label_type = raw.get(
            "label_type", None
        )  # Type of element triggering the event
        self.label = (
            raw.label.replace(b"\0", b" ")
            .strip(b" ")
            .decode(encoding=cfg.LABEL_ENCODING, errors="ignore")
        )  # Event Element Label. May be overridden locally

        event_map = copy(event_map[self.major])  # for inplace modifications

        # If 'sub' key is present, minor code will be used to obtain detailed information
        if "sub" in event_map:
            if self.minor in event_map["sub"]:
                sub = event_map["sub"][self.minor]
                if isinstance(sub, str):
                    sub = dict(message=sub)

                for k in sub:
                    if k == "message":
                        event_map[k] = (
                            "{}: {}".format(event_map[k], sub[k])
                            if k in event_map
                            else sub[k]
                        )
                    elif isinstance(sub[k], typing.List):  # for tags or other lists
                        event_map[k] = event_map.get(k, []) + sub[k]
                    else:
                        event_map[k] = sub[k]
            del event_map["sub"]
        # If key is not present, minor code contains the element ID
        else:
            self.id = self.minor

        callables = (
            k
            for k in event_map
            if isinstance(event_map[k], typing.Callable) and k != "hook_fn"
        )
        for k in callables:
            event_map[k] = event_map[k](self, self.label_provider)

        self.level = event_map.get("level", self.level)
        self.type = event_map.get("type", self.type)
        self._message_tpl = event_map.get("message", self._message_tpl)
        self.change = event_map.get("change", self.change)

        # ## Support parsing change values as messages
        # for k in self.change:
        #     if isinstance(self.change[k], str):
        #         self.change[k] = EventMessageFormatter().format(self.change[k], self)

        self.tags = event_map.get("tags", [])
        self.hook_fn = event_map.get("hook_fn")

        self.additional_data = {
            k: v
            for k, v in event_map.items()
            if k not in ["message"] and not hasattr(self, k)
        }

        # Set partition label
        if self.type == "partition":
            self.label = self.label_provider(self.type, self.partition)
            self.id = self.partition

    @property
    def name(self) -> str:
        key = self.partition if self.type == "partition" else self.minor

        name = self.label_provider(self.type, key)
        if name:
            return name

        return "-"


class ChangeEvent(Event):
    def __init__(self, change_object: Change, property_map, label_provider=None):
        if change_object.property not in property_map:
            raise AssertionError(
                "property %s not in property_map", change_object.property
            )

        super().__init__(label_provider=label_provider)

        self.timestamp = change_object.timestamp
        if change_object.type == "partition":
            self.partition = change_object.key
        else:
            self.partition = ""
        self.property = change_object.property

        self.value = change_object.new_value
        self.type = change_object.type
        self.label = change_object.key
        self.change = {self.property: self.value}

        property_map = copy(property_map[self.property])  # for inplace modifications
        callables = (
            k for k in property_map if isinstance(property_map[k], typing.Callable)
        )
        for k in callables:
            property_map[k] = property_map[k](self, self.label_provider)

        self.level = property_map.get("level", self.level)
        tpl = property_map.get("message", self._message_tpl)
        if isinstance(tpl, dict):
            if str(self.value) in tpl:
                self._message_tpl = tpl[str(self.value)]
        else:
            self._message_tpl = tpl

        self.tags = property_map.get("tags", [])


Notification = namedtuple("Notification", ["sender", "message", "level"])



from paradox.event import EventLevel

property_map = {
    "ac_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "AC Power failure", "False": "AC Power restored"},
    ),
    "aux_limit_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "{Type} {label} AUX Limit reached",
            "False": "{Type} {label} AUX Limit is OK",
        },
    ),
    "generated_alarm": dict(
        level=EventLevel.CRITICAL,  # Flag is set when alarm is triggered and unset when armed
        # next time
        message={
            "True": "{Type} {label} in alarm",
            "False": "{Type} {label} not in alarm",
        },
    ),
    "fire_alarm": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} in fire alarm",
            "False": "{Type} {label} not in fire alarm",
        },
    ),
    "alarm_duration_finished": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} alarm duration finished",
            "False": "{Type} {label} alarm duration active",
        },
    ),
    "alarm_in_memory": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} has alarm in memory",
            "False": "{Type} {label} alarm memory empty",
        },
    ),
    "arm": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} is armed",
            "False": "{Type} {label} is disarmed",
        },
    ),
    "arm_away": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} away arm active",
            "False": "{Type} {label} away arm inactive",
        },
    ),
    "arm_no_entry": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} no entry arm is active",
            "False": "{Type} {label} no entry arm is inactive",
        },
    ),
    "arm_stay": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} stay arm active",
            "False": "{Type} {label} stay arm inactive",
        },
    ),
    "all_zone_closed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} all zones closed",
            "False": "{Type} {label} not all zones closed",
        },
    ),
    "audible_alarm": dict(
        level=EventLevel.CRITICAL,  # Flag is set when partition is in alarm. Does not mean it
        # produces any sound at that moment. Unset when disarmed.
        message={
            "True": "{Type} {label} alarm is audible",
            "False": "{Type} {label} alarm is silent",
        },
    ),
    "auto_arm_reach": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} auto arm reached",
            "False": "{Type} {label} auto arm not reached",
        },
    ),
    "auto_arming_engaged": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} auto arming engaged",
            "False": "{Type} {label} auto arming is not engaged",
        },
    ),
    "bypass_ready": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} bypass is ready",
            "False": "{Type} {label} bypass is not ready",
        },
    ),
    "cancel_alarm_reporting_on_disarming": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],  #
        message={
            "True": "{Type} {label} alarm reporting on disarm is inactive",
            "False": "{Type} {label} alarm reporting on disarm is active",
        },
    ),
    "battery_failure_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "Battery failure trouble",
            "False": "Battery failure trouble recovered",
        },
    ),
    "follow_become_delay": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} follow becode delay is active",
            "False": "{Type} {label} follow becode delay is inactive",
        },
    ),
    "bell_absent_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "Bell is absent", "False": "Bell is present"},
    ),
    "bell_limit_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "Bell output overloaded",
            "False": "Bell output load is adequate",
        },
    ),
    "bypassed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} is bypassed",
            "False": "{Type} {label} is not bypassed",
        },
    ),
    "force_ready": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} force ready",
            "False": "{Type} {label} force not ready",
        },
    ),
    "inhibit_ready": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} inhibit ready",
            "False": "{Type} {label} inhibit not ready",
        },
    ),
    "intellizone_engage": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} intellizone engadged",
            "False": "{Type} {label} intellizone not engadged",
        },
    ),
    "lockout": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "{Type} {label} lockout",
            "False": "{Type} {label} lockout inactive",
        },
    ),
    "no_movement_delay_end": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} no movement delay ended",
            "False": "{Type} {label} no movement delay is active",
        },
    ),
    "entry_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} in entry delay",
            "False": "{Type} {label} not in entry delay",
        },
    ),
    "entry_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} entry delay is finished",
            "False": "{Type} {label} entry delay is not finished",
        },
    ),
    "exit_delay": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} in exit delay",
            "False": "{Type} {label} not in exit delay",
        },
    ),
    "exit_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} exit delay is finished",
            "False": "{Type} {label} exit delay is not finished",
        },
    ),
    "fire_delay_in_progress": dict(
        level=EventLevel.INFO,  #
        message={
            "True": "{Type} {label} is in fire delay",
            "False": "{Type} {label} is in fire delay",
        },
    ),
    "fire_delay_end": dict(
        level=EventLevel.INFO,  #
        message={
            "True": "{Type} {label} fire delay ended",
            "False": "{Type} {label} fire delay not ended",
        },
    ),
    "intellizone_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} intellizone delay finished",
            "False": "{Type} {label} intellizone delay is active",
        },
    ),
    "module_tamper_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "Module tampered", "False": "Module tamper cleared"},
    ),
    "open": dict(
        level=EventLevel.DEBUG,
        message={"True": "Zone {label} open", "False": "Zone {label} closed"},
    ),
    "panic_alarm": dict(
        level=EventLevel.CRITICAL,
        tags=["alarm"],
        message={
            "True": "{Type} {label} panic alarm active",
            "False": "{Type} {label} panic alarm inactive",
        },
    ),
    "partition_recently_close": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} partition recently close active",
            "False": "{Type} {label} partition recently close inactive",
        },
    ),
    "programming": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} programming active",
            "False": "{Type} {label} programming inactive",
        },
    ),
    "ready": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} ready", "False": "{Type} {label} not ready"},
    ),
    "remote_arming": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} remote arming active",
            "False": "{Type} {label} remote arming inactive",
        },
    ),
    "police_code_delay": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} police code delay active",
            "False": "{Type} {label} police code delay inactive",
        },
    ),
    "silent_alarm": dict(
        level=EventLevel.CRITICAL,
        message={
            "True": "{Type} {label} silent alarm is active",
            "False": "{Type} {label} silent alarm is inactive",
        },
    ),
    "stay_arming_auto": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} auto stay arming active",
            "False": "{Type} {label} auto stay arming inactive",
        },
    ),
    "stay_instant_ready": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} stay instant ready",
            "False": "{Type} {label} stay instant not ready",
        },
    ),
    "supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "{Type} {label} supervision trouble",
            "False": "{Type} {label} supervision OK",
        },
    ),
    "tamper": dict(
        level=EventLevel.WARN,
        tags=["trouble"],
        message={
            "True": "{Type} {label} tampered",
            "False": "{Type} {label} tamper cleared",
        },
    ),
    "time_to_refresh_zone_status": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} time to refresh zone status",
            "False": "{Type} {label} time to refresh zone status not came",
        },
    ),
    "trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "{Type} {label} has trouble",
            "False": "No trouble for {Type} {label}",
        },
    ),
    "tx_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} tx delay finished",
            "False": "{Type} {label} tx delay not finished",
        },
    ),
    "voice_arming": dict(
        level=EventLevel.DEBUG,
        message={
            "True": "{Type} {label} voice arming active",
            "False": "{Type} {label} voice arming inactive",
        },
    ),
    "was_in_alarm": dict(
        level=EventLevel.INFO,  # Flag is set when partition is in alarm, unset when disarmed
        message={
            "True": "{Type} {label} was in alarm",
            "False": "{Type} {label} was not in alarm",
        },
    ),
    "zone_bypassed": dict(
        level=EventLevel.INFO,
        message={
            "True": "{Type} {label} a zone is bypassed",
            "False": "{Type} {label} no zones bypassed",
        },
    ),
    "zone_fire_loop_trouble": dict(
        level=EventLevel.WARN,
        tags=["trouble"],
        message={
            "True": "{Type} {label} zone fire loop trouble",
            "False": "{Type} {label} zone fire loop OK",
        },
    ),
    "zone_supervision_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "Zone with supervision trouble",
            "False": "Zone supervision is OK",
        },
    ),
    "zone_tamper_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "Zone tampered", "False": "No Zone is tampered"},
    ),
    "zone_low_battery_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "zone low battery trouble",
            "False": "zone low battery trouble recovered",
        },
    ),
    "bus_com_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "bus com trouble", "False": "bus com OK"},
    ),
    "bus_global_fail": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "bus global fail", "False": "bus OK"},
    ),
    "bus_overload_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "bus overload trouble", "False": "bus load OK"},
    ),
    "com_pc_trouble": dict(
        level=EventLevel.INFO,
        tags=["trouble"],
        message={"True": "pc communication trouble", "False": "pc communication OK"},
    ),
    "dialer_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "dialer trouble active", "False": "dialer trouble inactive"},
    ),
    "fail_tel_1_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "fail tel 1 trouble active",
            "False": "fail tel 1 trouble inactive",
        },
    ),
    "fail_tel_2_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "fail tel 2 trouble active",
            "False": "fail tel 2 trouble inactive",
        },
    ),
    "fail_tel_3_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "fail tel 3 trouble active",
            "False": "fail tel 3 trouble inactive",
        },
    ),
    "fail_tel_4_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "fail tel 4 trouble active",
            "False": "fail tel 4 trouble inactive",
        },
    ),
    "mdl_com_error": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "module communication error active",
            "False": "module communication error",
        },
    ),
    "missing_keypad_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "missing keypad", "False": "keypad present"},
    ),
    "missing_module_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "missing a module", "False": "all modules present"},
    ),
    "module_ac_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module AC trouble", "False": "all module AC OK"},
    ),
    "module_aux_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module aux trouble", "False": "module aux OK"},
    ),
    "module_battery_fail": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module battery fail", "False": "module battery OK"},
    ),
    "module_fail_to_com_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "module fail to communicate trouble",
            "False": "module communication OK",
        },
    ),
    "module_printer_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module printer trouble", "False": "module printer OK"},
    ),
    "module_rom_error_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={
            "True": "module rom error trouble",
            "False": "module rom error trouble recovered",
        },
    ),
    "module_tlm_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module tlm trouble", "False": "module tlm trouble recovered"},
    ),
    "module_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "module trouble", "False": "module trouble recovered"},
    ),
    "rom_error_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "rom error trouble", "False": "rom error trouble recovered"},
    ),
    "safety_mismatch_trouble": dict(
        level=EventLevel.INFO,
        tags=["trouble"],
        message={
            "True": "safety mismatch trouble",
            "False": "safety mismatch trouble recovered",
        },
    ),
    "system_trouble": dict(
        level=EventLevel.INFO,
        tags=["trouble"],
        message={"True": "system trouble", "False": "system trouble recovered"},
    ),
    "time_lost_trouble": dict(
        level=EventLevel.INFO,
        tags=["trouble"],
        message={"True": "time lost trouble", "False": "time lost trouble recovered"},
    ),
    "tlm_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "tlm trouble", "False": "tlm trouble recovered"},
    ),
    "zone_fault_trouble": dict(
        level=EventLevel.CRITICAL,
        tags=["trouble"],
        message={"True": "zone fault trouble", "False": "zone fault trouble recovered"},
    ),
    "activated_entry_delay": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "True": "{Type} {label} activated entry delay active",
            "False": "{Type} {label} activated entry delay inactive",
        },
    ),
    "activated_intellizone_delay": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "True": "{Type} {label} activated intellizone delay active",
            "False": "{Type} {label} activated intellizone delay inactive",
        },
    ),
    "low_battery": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "True": "{Type} {label} low battery",
            "False": "{Type} {label} battery OK",
        },
    ),
    "presently_in_alarm": dict(
        level=EventLevel.INFO,
        tags=[],  # Flag is set when zone is in alarm
        message={
            "True": "{Type} {label} triggered alarm",
            "False": "{Type} {label} alarm restored",
        },
    ),
    "shut_down": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "True": "{Type} {label} shut down active",
            "False": "{Type} {label} shut down inactive",
        },
    ),
    "tx_delay": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "True": "{Type} {label} tx delay active",
            "False": "{Type} {label} tx delay inactive",
        },
    ),
    "time": dict(
        level=EventLevel.DEBUG, tags=["system", "time"], message="Panel time is {value}"
    ),
    # ---------------------
    # Synthetic properties
    # ---------------------
    "current_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "triggered": "{Type} {label} triggered alarm",
            "pending": "{Type} {label} arming pending",
            "armed_home": "{Type} {label} armed stay",
            "armed_away": "{Type} {label} armed away",
            "armed_night": "{Type} {label} armed night",
            "disarmed": "{Type} {label} disarmed",
        },
    ),
    "target_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "armed_home": "{Type} {label} arm stay",
            "armed_away": "{Type} {label} arm away",
            "armed_night": "{Type} {label} arm night",
            "disarmed": "{Type} {label} disarm",
        },
    ),
}



import logging

from paradox.config import config as cfg
from paradox.event import Event, EventLevel, Notification
from paradox.interfaces import ThreadQueueInterface
from paradox.lib import ps
from paradox.lib.event_filter import (EventFilter, EventTagFilter,
                                      LiveEventRegexpFilter)

logger = logging.getLogger("PAI").getChild(__name__)


class AbstractTextInterface(ThreadQueueInterface):
    """Interface Class using any Text interface"""

    def __init__(self, alarm, event_filter: EventFilter, min_level=EventLevel.INFO):
        super().__init__(alarm)

        self.event_filter = event_filter

        self.min_level = min_level
        self.alarm = alarm

    def stop(self):
        super().stop()

    def _run(self):
        super(AbstractTextInterface, self)._run()

        ps.subscribe(self.handle_panel_event, "events")
        ps.subscribe(self.handle_notify, "notifications")

    def send_message(self, message: str, level: EventLevel):
        pass

    def notification_filter(self, notification: Notification):
        return notification.level >= self.min_level and notification.sender != self.name

    def handle_notify(self, notification: Notification):
        if self.notification_filter(notification):
            self.send_message(notification.message, notification.level)

    def handle_panel_event(self, event: Event):
        if self.event_filter.match(event):
            self.send_message(event.message, event.level)

    async def handle_command(self, message_raw):
        message = cfg.COMMAND_ALIAS.get(message_raw, message_raw)

        tokens = message.split(" ")

        if len(tokens) != 3:
            m = "Invalid: {}".format(message_raw)
            logger.warning(m)
            return m

        if self.alarm is None:
            m = "No alarm registered"
            logger.error(m)
            return m

        element_type = tokens[0].lower()
        element = tokens[1]
        command = self.normalize_payload(tokens[2].lower())

        # Process a Zone Command
        if element_type == "zone":
            if not await self.alarm.control_zone(element, command):
                m = "Zone command error: {}={}".format(element, command)
                logger.warning(m)
                return m

        # Process a Partition Command
        elif element_type == "partition":
            if not await self.alarm.control_partition(element, command):
                m = "Partition command error: {}={}".format(element, command)
                logger.warning(m)
                return m

        # Process an Output Command
        elif element_type == "output":
            if not await self.alarm.control_output(element, command):
                m = "Output command error: {}={}".format(element, command)
                logger.warning(m)
                return m
        else:
            m = "Invalid control element: {}".format(element)
            logger.error(m)
            return m

        logger.info("OK: {}".format(message_raw))
        return "OK"

    # TODO: Remove this (to panels?)
    @staticmethod
    def normalize_payload(message):
        message = message.strip().lower()

        if message in ["true", "on", "1", "enable"]:
            return "on"
        elif message in ["false", "off", "0", "disable"]:
            return "off"
        elif message in [
            "pulse",
            "arm",
            "disarm",
            "arm_stay",
            "arm_sleep",
            "bypass",
            "clear_bypass",
        ]:
            return message

        return None


class ConfiguredAbstractTextInterface(AbstractTextInterface):
    def __init__(
        self, alarm, EVENT_FILTERS, ALLOW_EVENTS, IGNORE_EVENTS, MIN_EVENT_LEVEL
    ):
        if EVENT_FILTERS and (ALLOW_EVENTS or IGNORE_EVENTS):
            raise AssertionError(
                "You can not use *_EVENT_FILTERS and *_ALLOW_EVENTS+*_IGNORE_EVENTS simultaneously"
            )

        min_level = EventLevel.from_name(MIN_EVENT_LEVEL)
        if ALLOW_EVENTS or IGNORE_EVENTS:  # Use if defined, else use TAGS as default
            logger.debug("Using REGEXP Filter")
            event_filter = LiveEventRegexpFilter(ALLOW_EVENTS, IGNORE_EVENTS, min_level)
        else:
            logger.debug("Using Tag Filter")
            event_filter = EventTagFilter(EVENT_FILTERS, min_level)

        super().__init__(alarm, event_filter=event_filter, min_level=min_level)



import logging
from collections import defaultdict
from typing import Callable

from paradox.config import config as cfg
from paradox.event import Change
from paradox.lib import ps

from .element_type_container import ElementTypeContainer

logger = logging.getLogger("PAI").getChild(__name__)


class MemoryStorage:
    def __init__(self):
        self.data = defaultdict(ElementTypeContainer)

    def get_container(self, container_name) -> ElementTypeContainer:
        return self.data[container_name]

    def get_container_object(self, container_name, key, create_if_missing=False):
        c = self.get_container(container_name)
        el = c.get(key)
        if create_if_missing and el is None:
            el = c[key] = {"key": key}

        return el

    def update_container_object(self, container_name: str, key: str, changes: dict):
        assert container_name is not None
        assert key is not None
        assert isinstance(changes, dict)
        if not changes:  # Has at least one element
            return

        # logger.debug('update_properties %s/%s=%s', container_name, key, change)
        element = self.get_container_object(container_name, key, create_if_missing=True)
        object_key = element["key"]

        # Publish changes and update state
        for property_name, property_value in changes.items():

            if not isinstance(property_name, str):
                logger.debug(
                    "Invalid property name ({}/{}/{}) type: {}".format(
                        container_name, object_key, property_name, type(property_name)
                    )
                )
                continue
            if property_name.startswith("_"):  # skip private properties
                continue

            old = element.get(property_name)

            if isinstance(
                property_value, Callable
            ):  # function to make new value from the old one
                try:
                    property_value = property_value(old)
                except:
                    logger.exception(
                        'Exception caught during property "%s" convert. Ignoring',
                        property_name,
                    )
                    continue

            # Standard processing of changes
            if property_name in element:
                if old == property_value and not cfg.PUSH_UPDATE_WITHOUT_CHANGE:
                    continue
                element[property_name] = property_value
            else:
                element[property_name] = property_value  # Initial value, do not notify
                # suppress = 'trouble' not in property_name

            change_object = Change(
                container_name, object_key, property_name, property_value, old
            )
            logger.debug(change_object)
            ps.sendChange(change_object)



# -*- coding: utf-8 -*-


import logging
import os
import stat
import typing

import serial_asyncio
from serial import SerialException

from ..exceptions import SerialConnectionOpenFailed
from .connection import Connection
from .handler import ConnectionHandler
from .protocols import SerialConnectionProtocol

logger = logging.getLogger("PAI").getChild(__name__)


class SerialCommunication(Connection, ConnectionHandler):
    def __init__(self, port, baud=9600):
        super().__init__()
        self.port_path = port
        self.baud = baud
        self.connected_future = None

    def on_connection_loss(self):
        logger.error("Connection to panel was lost")
        self.connected = False
        if not self.connected_future.done():
            self.connected_future.set_result(self.connected)

    def on_connection(self):
        logger.info("Serial port open")
        self.connected = True
        if not self.connected_future.done():
            self.connected_future.set_result(self.connected)

    def open_timeout(self):
        if self.connected_future.done():
            return

        logger.error("Serial Port Timeout")
        self.connected = False
        self.connected_future.set_result(self.connected)

    def make_protocol(self):
        return SerialConnectionProtocol(self)

    async def connect(self) -> bool:
        logger.info(f"Connecting to serial port {self.port_path}")

        if not os.access(self.port_path, mode=os.R_OK | os.W_OK):
            logger.info(f"{self.port_path} is not readable/writable. Trying to fix...")
            try:
                os.chmod(
                    self.port_path,
                    stat.S_IRUSR
                    | stat.S_IWUSR
                    | stat.S_IRGRP
                    | stat.S_IWGRP
                    | stat.S_IROTH
                    | stat.S_IWOTH,
                )
                logger.info(f"File {self.port_path} permissions changed")
            except OSError:
                logger.error(f"Failed to update file {self.port_path} permissions")
                return False

        self.connected_future = self.loop.create_future()
        open_timeout_handler = self.loop.call_later(5, self.open_timeout)

        try:
            _, self._protocol = await serial_asyncio.create_serial_connection(
                self.loop, self.make_protocol, self.port_path, self.baud
            )

            return await self.connected_future
        except SerialException as e:
            self.connected_future.cancel()
            raise SerialConnectionOpenFailed(
                "Connection to serial port failed"
            ) from e  # PAICriticalException
        except:
            logger.exception("Unable to connect to Serial")
        finally:
            open_timeout_handler.cancel()

        return False



from construct import Container

from .panel import Panel


def create_panel(core, start_communication_response: Container = None) -> Panel:
    product_id = (
        start_communication_response.fields.value.product_id
        if start_communication_response
        else None
    )
    if product_id is None:
        return Panel(core, True)
    elif product_id == "DIGIPLEX_EVO_48":
        from . import evo

        return evo.Panel_EVO48(core, start_communication_response, True)
    elif product_id == "DIGIPLEX_EVO_96":
        from . import evo

        return evo.Panel_EVO96(core, start_communication_response, True)
    elif product_id == "DIGIPLEX_EVO_192":
        from . import evo

        return evo.Panel_EVO192(core, start_communication_response, True)
    elif product_id == "DIGIPLEX_EVO_HD":
        from . import evo

        return evo.Panel_EVOHD(core, start_communication_response, True)
    elif product_id in [
        "SPECTRA_SP4000",
        "SPECTRA_SP5500",
        "SPECTRA_SP6000",
        "SPECTRA_SP7000",
        "SPECTRA_SP65",
        "MAGELLAN_MG5000",
        "MAGELLAN_MG5050",
        "MAGELLAN_MG5075",
    ]:
        from . import spectra_magellan

        return spectra_magellan.Panel(core, start_communication_response, False)
    else:
        raise NotImplementedError(
            "We are not sure what panel you have (product_id: {}). \
            Please create an issue. Maybe we can help you.".format(
                str(product_id)
            )
        )



# -*- coding: utf-8 -*-

import asyncio
import binascii
import inspect
import logging
import typing

from construct import ChecksumError, Construct, Container

from paradox.config import config as cfg
from paradox.exceptions import AuthenticationFailed, StatusRequestException

from ..panel import Panel as PanelBase
from . import parsers
from .event import event_map
from .property import property_map

logger = logging.getLogger("PAI").getChild(__name__)

PARTITION_ACTIONS = dict(
    arm=0x04,
    disarm=0x05,
    arm_stay=0x01,
    arm_force=0x02,
    arm_sleep=0x03,
    arm_stay_stayd=0x06,
    arm_sleep_stay=0x07,
    disarm_all=0x08,
)
ZONE_ACTIONS = dict(bypass=0x10, clear_bypass=0x10)
PGM_ACTIONS = dict(on_override=0x30, off_override=0x31, on=0x32, off=0x33, pulse=0)


class Panel(PanelBase):

    event_map = event_map
    property_map = property_map
    max_eeprom_response_data_length = 32
    status_request_addresses = parsers.RAMDataParserMap.keys()

    mem_map = {
        "status_base1": 0x8000,
        "status_base2": 0x1FE0,
        "definitions": {"zone": {"addresses": [range(0x730, 0x800, 0x03)]}},
        "labels": {
            "zone": {"label_offset": 0, "addresses": [range(0x010, 0x210, 0x10)]},
            "pgm": {
                "label_offset": 0,
                "addresses": [range(0x210, 0x310, 0x10)],
                "template": {"on": False, "pulse": False},
            },
            "partition": {"label_offset": 0, "addresses": [range(0x310, 0x330, 0x10)]},
            "user": {"label_offset": 0, "addresses": [range(0x330, 0x530, 0x10)]},
            "module": {"label_offset": 0, "addresses": [range(0x530, 0x620, 0x10)]},
            "repeater": {"label_offset": 0, "addresses": [range(0x620, 0x640, 0x10)]},
            "keypad": {"label_offset": 0, "addresses": [range(0x640, 0x6C0, 0x10)]},
            "site": {"label_offset": 0, "addresses": [range(0x6C0, 0x6D0, 0x10)]},
            "siren": {"label_offset": 0, "addresses": [range(0x6D0, 0x700, 0x10)]},
        },
    }

    def __init__(
        self, core, start_communication_response, variable_message_length=True
    ):
        super(Panel, self).__init__(core, variable_message_length)

        self.settings = start_communication_response.fields.value

    async def dump_memory(self, file, memory_type):
        """
        Dumps EEPROM and RAM memory to files
        :return:
        """
        if memory_type == "ram":
            await self.dump_memory_to_file(file, range(0, 9), ram=True)
        elif memory_type == "eeprom":
            await self.dump_memory_to_file(file, range(0, 0x0FFF, 32))
        else:
            raise AttributeError(f"Unknown memory type: {memory_type}")

    async def dump_memory_to_file(self, file, range_, ram=False):
        mem_type = "RAM" if ram else "EEPROM"
        logger.info("Dump " + mem_type)

        for address in range_:
            if ram:
                args = dict(address=address + self.mem_map["status_base1"])
            else:
                args = dict(address=address)

            logger.info("Dumping %s: address %x" % (mem_type, address))

            reply = await self.core.send_wait(
                parsers.ReadEEPROM,
                args,
                reply_expected=lambda m: m.fields.value.po.command == 0x05
                and m.fields.value.address == address,
            )

            if reply is None:
                logger.error("Could not read %s: address %x" % (mem_type, address))
                return

            data = reply.fields.value.data

            file.write(data)

    def get_message(self, name: str) -> Construct:
        try:
            clsmembers = dict(inspect.getmembers(parsers))
            if name in clsmembers:
                return clsmembers[name]
        except ResourceWarning:
            pass

        return super(Panel, self).get_message(name)

    def parse_message(
        self, message: bytes, direction="topanel"
    ) -> typing.Optional[Container]:
        try:
            if message is None or len(message) == 0:
                return None

            parent_parsed = super(Panel, self).parse_message(message, direction)
            if parent_parsed:
                return parent_parsed

            if direction == "topanel":
                if message[0] == 0x70 and message[-5] != 0:
                    return parsers.CloseConnection.parse(message)
                elif message[0] == 0x00:
                    return parsers.InitializeCommunication.parse(message)
                elif message[0] == 0x30:
                    return parsers.SetTimeDate.parse(message)
                elif message[0] == 0x40:
                    return parsers.PerformAction.parse(message)
                elif message[0] == 0x50 and message[2] < 0x80:
                    return parsers.ReadEEPROM.parse(message)

            else:
                if message[0] == 0x10:
                    return parsers.InitializeCommunicationResponse.parse(message)
                elif message[0] >> 4 == 0x7 and message[-5] == 0:
                    return parsers.ErrorMessage.parse(message)
                elif message[0] >> 4 == 0x03:
                    return parsers.SetTimeDateResponse.parse(message)
                elif message[0] >> 4 == 0x04:
                    return parsers.PerformActionResponse.parse(message)
                elif message[0] >> 4 == 0x05 and message[2] == 0x80:
                    return parsers.ReadStatusResponse.parse(message)
                elif message[0] >> 4 == 0x05 and message[2] < 0x80:
                    return parsers.ReadEEPROMResponse.parse(message)

                #        elif message[0] == 0x60 and message[2] < 0x80:
                #            return WriteEEPROM.parse(message)
                #        elif message[0] >> 4 == 0x06 and message[2] < 0x80:
                #            return WriteEEPROMResponse.parse(message)
                elif message[0] >> 4 == 0x0E:
                    return parsers.LiveEvent.parse(message)

        except ChecksumError as e:
            logger.error(
                "ChecksumError %s, message: %s" % (str(e), binascii.hexlify(message))
            )
        except:
            logger.exception(
                "Exception parsing message: %s" % (binascii.hexlify(message))
            )
        return None

    async def initialize_communication(self, password):
        encoded_password = self.encode_password(password)

        args = dict(
            product_id=self.settings.product_id,
            firmware=self.settings.firmware,
            panel_id=self.settings.panel_id,
            pc_password=encoded_password,
            user_code=0x000000,
            _not_used1=0x19,
            source_id=0x02,
        )

        logger.info("Installer login")
        reply = await self.core.send_wait(
            parsers.InitializeCommunication,
            args=args,
            reply_expected=[0x10, 0x70, 0x00],
        )

        if reply is None:
            logger.error("Installer login failed")
            return False

        if reply.fields.value.po.command == 0x10:
            logger.info("Authentication Success")
            return True
        elif (
            reply.fields.value.po.command == 0x70
            or reply.fields.value.po.command == 0x00
        ):
            logger.error("Authentication Failed. Wrong Password?")
            raise AuthenticationFailed("Wrong PASSWORD")

    @staticmethod
    def _request_status_reply_check(message: Container, address: int):
        mvars = message.fields.value

        if mvars.po.command == 0x05 and mvars.address == address:
            return True

        return False

    async def request_status(self, i: int):
        args = dict(address=self.mem_map["status_base1"] + i)
        reply = await self.core.send_wait(
            parsers.ReadEEPROM,
            args,
            reply_expected=lambda m: self._request_status_reply_check(m, i),
        )
        if reply is not None:
            logger.debug("Received status response: %d" % i)
            return self.handle_status(reply, parsers.RAMDataParserMap)
        else:
            raise StatusRequestException("No reply to status request: %d" % i)

    async def control_zones(self, zones: list, command: str) -> bool:
        """
        Control zones
        :param list zones: a list of zones
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in ZONE_ACTIONS:
            return False

        accepted = False

        for zone in zones:
            args = dict(action=ZONE_ACTIONS[command], argument=(zone - 1))
            reply = await self.core.send_wait(
                parsers.PerformAction, args, reply_expected=0x04
            )

            if reply is not None:
                accepted = True

        return accepted

    async def control_partitions(self, partitions: list, command: str) -> bool:
        """
        Control Partitions
        :param list partitions: a list of partitions
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in PARTITION_ACTIONS:
            return False

        accepted = False

        for partition in partitions:
            args = dict(action=PARTITION_ACTIONS[command], argument=(partition - 1))
            reply = await self.core.send_wait(
                parsers.PerformAction, args, reply_expected=0x04
            )

            if reply is not None:
                accepted = True

        return accepted

    async def control_outputs(self, outputs, command) -> bool:
        """
        Control PGM
        :param list outputs: a list of pgms
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in PGM_ACTIONS:
            return False

        accepted = False

        for output in outputs:
            if command == "pulse":
                args = dict(action=PGM_ACTIONS["on"], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x04
                )
                if reply is None:
                    continue

                await asyncio.sleep(cfg.OUTPUT_PULSE_DURATION)
                args = dict(action=PGM_ACTIONS["off"], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x04
                )
                if reply is not None:
                    accepted = True
            else:
                args = dict(action=PGM_ACTIONS[command], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x04
                )
                if reply is not None:
                    accepted = True

        return accepted

    async def send_panic(self, partitions, panic_type, user_id):
        accepted = False

        args = {"partitions": partitions, "panic_type": panic_type, "user_id": user_id}

        reply = await self.core.send_wait(
            parsers.SendPanicAction, args, reply_expected=0x04
        )

        if reply is not None:
            accepted = True

        return accepted



# -*- coding: utf-8 -*-

import asyncio
import logging
import threading

logger = logging.getLogger("PAI").getChild(__name__)


class Interface:
    def __init__(self, alarm):
        super().__init__()  # yes it is required!
        self.name = self.__class__.__name__
        self.alarm = alarm

    def start(self):
        pass


class AsyncInterface(Interface):
    def __init__(self, alarm):
        super().__init__(alarm)

        self._loop = asyncio.get_event_loop()
        self._running_task = None  # type: asyncio.Task

    def start(self):
        self._running_task = self._loop.create_task(self.run())

    def stop(self):
        if self._running_task and not self._running_task.done():
            self._running_task.cancel()

    async def run(self):
        pass


class ThreadQueueInterface(threading.Thread, Interface):
    def __init__(self, alarm):
        super().__init__(name=self.__class__.__name__)

        self.alarm = alarm

        self.stop_running = threading.Event()
        self.stop_running.clear()

    def stop(self):
        self.stop_running.set()
        if threading.current_thread() != self:
            self.join()
            logger.debug("Interface %s thread stopped", self.name)

    def run(self):
        try:
            self._run()
        except (KeyboardInterrupt, SystemExit):
            logger.debug("Interface loop stopping")
            self.stop()
        except:
            logger.exception("Interface loop")

    def _run(self):
        logger.info("Starting %s Interface", self.name)



from construct import (BitsInteger, BitStruct, Byte, Bytes, Const, Default,
                       Enum, ExprAdapter, Flag, Int8ub, Int8ul, Int16ub,
                       Int16ul, Int32ub, Nibble, Padding, RawCopy, Struct,
                       obj_)

from ..common import CommunicationSourceIDEnum, PacketChecksum, ProductIdEnum
from .adapters import (DateAdapter, ModuleSerialAdapter,
                       PartitionStatusAdapter, SignalStrengthAdapter,
                       StatusAdapter, ZoneStatusAdapter)

InitializeCommunication = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x00, Int8ub)),
            "_not_used0" / Padding(3),
            "product_id" / ProductIdEnum,
            "firmware"
            / Struct("version" / Int8ub, "revision" / Int8ub, "build" / Int8ub),
            "panel_id" / Int16ub,
            "pc_password" / Default(Bytes(2), b"0000"),
            "_not_used1" / Bytes(1),
            "source_method"
            / Default(
                Enum(Int8ub, Winload_Connection=0x00, NEware_Connection=0x55), 0x00
            ),
            "user_code" / Default(Int32ub, 0x00000000),
            "_not_used2" / Padding(15),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

InitializeCommunicationResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x10, Int8ub)),
            "neware_connection" / Int16ub,
            "user_id_low" / Int8ub,
            "partition_rights"
            / BitStruct(
                "_not_used" / BitsInteger(6), "partition_2" / Flag, "partition_1" / Flag
            ),
            "_not_used0" / Padding(31),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

PanelStatus = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x50, Int8ub)),
            "_not_used0" / Default(Int8ub, 0x00),
            "validation" / Default(Int8ub, 0x00),
            "status_request" / Default(Int8ub, 0x00),
            "_not_used1" / Padding(29),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    Padding(31),
    "checksum" / PacketChecksum(Bytes(1)),
)

RAMDataParserMap = {
    0: Struct(
        "troubles"
        / BitStruct(
            "timer_loss_trouble" / Flag,
            "fire_loop_trouble" / Flag,
            "module_tamper_trouble" / Flag,
            "zone_tamper_trouble" / Flag,
            "communication_trouble" / Flag,
            "bell_trouble" / Flag,
            "power_trouble" / Flag,
            "rf_low_battery_trouble" / Flag,
            "rf_interference_trouble" / Flag,
            "_not_used0" / BitsInteger(5),
            "module_supervision_trouble" / Flag,
            "zone_supervision_trouble" / Flag,
            "_not_used1" / BitsInteger(1),
            "wireless_repeater_battery_trouble" / Flag,
            "wireless_repeater_ac_loss_trouble" / Flag,
            "wireless_keypad_battery_trouble" / Flag,
            "wireless_keypad_ac_trouble" / Flag,
            "auxiliary_output_overload_trouble" / Flag,
            "ac_failure_trouble" / Flag,
            "low_battery_trouble" / Flag,
            "_not_used2" / BitsInteger(6),
            "bell_output_overload_trouble" / Flag,
            "bell_output_disconnected_trouble" / Flag,
            "_not_used3" / BitsInteger(2),
            "computer_fail_to_communicate_trouble" / Flag,
            "voice_fail_to_communicate_trouble" / Flag,
            "pager_fail_to_communicate_trouble" / Flag,
            "central_2_reporting_ftc_indicator_trouble" / Flag,
            "central_1_reporting_ftc_indicator_trouble" / Flag,
            "telephone_line" / Flag,
        ),
        "system"
        / Struct(
            "date" / Struct("time" / DateAdapter(Bytes(6))),
            "power"
            / Struct(
                "vdc"
                / ExprAdapter(
                    Byte, lambda obj, ctx: round(obj * (20.3 - 1.4) / 255.0 + 1.4, 1), 0
                ),
                "battery"
                / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
                "dc"
                / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
            ),
            "rf" / Struct("noise_floor" / Int8ub,),
        ),
        "zone_open" / StatusAdapter(Bytes(4)),
        "zone_tamper" / StatusAdapter(Bytes(4)),
        "pgm_tamper" / StatusAdapter(Bytes(2)),
        "module_tamper" / StatusAdapter(Bytes(2)),
        "zone_fire" / StatusAdapter(Bytes(4)),
        "_not_used0" / Int8ub,
    ),
    1: Struct(
        "zone_rf_supervision_trouble" / StatusAdapter(Bytes(4)),
        "pgm_supervision_trouble" / StatusAdapter(Bytes(2)),
        "module_supervision_trouble" / StatusAdapter(Bytes(2)),
        "repeater_supervision_trouble" / StatusAdapter(Bytes(1)),
        "zone_rf_low_battery_trouble" / StatusAdapter(Bytes(4)),
        "partition_status" / PartitionStatusAdapter(Bytes(8)),
        "repeater_ac_loss_trouble" / StatusAdapter(Bytes(1)),
        "repeater_battery_failure_trouble" / StatusAdapter(Bytes(1)),
        "keypad_ac_loss_trouble" / StatusAdapter(Bytes(1)),
        "keypad_battery_failure_trouble" / StatusAdapter(Bytes(1)),
        "keypad_supervision_failure_trouble" / StatusAdapter(Bytes(1)),
        "_not_used0" / Padding(6),
    ),
    2: Struct("zone_status" / ZoneStatusAdapter(Bytes(32))),
    3: Struct("zone_signal_strength" / SignalStrengthAdapter(Bytes(32))),
    4: Struct(
        "pgm_signal_strength" / SignalStrengthAdapter(Bytes(16)),
        "repeater_signal_strength" / SignalStrengthAdapter(Bytes(2)),
        "keypad_signal_strength" / SignalStrengthAdapter(Bytes(8)),
        "_not_used1" / Padding(6),
    ),
    5: Struct("zone_exit_delay" / StatusAdapter(Bytes(4)), "_not_used0" / Padding(28)),
}

LiveEvent = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0xE, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "time" / DateAdapter(Bytes(6)),
            "event" / Struct("major" / Int8ub, "minor" / Int8ub),
            "partition" / ExprAdapter(Byte, obj_ + 1, obj_ - 1),
            "module_serial" / ModuleSerialAdapter(Bytes(4)),
            "unknown0" / Bytes(1),
            "label" / Bytes(16),
            "unknown1" / Bytes(1),
            "reserved0" / Bytes(4),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

CloseConnection = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x70, Int8ub)),
            "_not_used0" / Const(0, Int8ub),
            "validation_byte" / Default(Int8ub, 0),
            "_not_used1" / Padding(29),
            "message"
            / Default(
                Enum(Int8ub, authentication_failed=0x12, panel_will_disconnect=0x05),
                0x05,
            ),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ErrorMessage = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x7, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Default(Int8ub, 0),
            "message"
            / Enum(
                Int8ub,
                requested_command_failed=0x00,
                invalid_user_code=0x01,
                partition_in_code_lockout=0x02,
                panel_will_disconnect=0x05,
                panel_not_connected=0x10,
                panel_already_connected=0x11,
                invalid_pc_password=0x12,
                winload_on_phone_line=0x13,
                invalid_module_address=0x14,
                cannot_write_in_ram=0x15,
                upgrade_request_fail=0x16,
                record_number_out_of_range=0x17,
                invalid_record_type=0x19,
                multibus_not_supported=0x1A,
                incorrect_number_of_users=0x1B,
                invalid_label_number=0x1C,
            ),
            "_not_used1" / Padding(33),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ReadEEPROM = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x50, Int8ub)),
            "_not_used0" / Padding(1),
            "address" / Default(Int16ub, 0),
            "_not_used1" / Padding(29),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ReadEEPROMResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x5, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Padding(1),
            "address" / Int16ub,
            "data" / Bytes(32),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ReadStatusResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x5, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Padding(1),
            "validation" / Const(0x80, Int8ub),
            "address" / Int8ub,
            "data" / Bytes(32),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

# noinspection PyUnresolvedReferences
SetTimeDate = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x30, Int8ub)),
            "_not_used0" / Padding(3),
            "century" / Int8ub,
            "year" / Int8ub,
            "month" / Int8ub,
            "day" / Int8ub,
            "hour" / Int8ub,
            "minute" / Int8ub,
            "_not_used1" / Padding(23),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

SetTimeDateResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x3, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Padding(35),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

PerformAction = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "_not_used0" / Padding(1),
            "action"
            / Enum(
                Int8ub,
                Stay_Arm=0x01,
                Stay_Arm1=0x02,
                Sleep_Arm=0x03,
                Full_Arm=0x04,
                Disarm=0x05,
                Stay_Arm_StayD=0x06,
                Sleep_Arm_StayD=0x07,
                Disarm_Both_Disable_StayD=0x08,
                Bypass=0x10,
                # Beep=0x10,
                Clear_Alarm_In_Memory=0x1B,
                PGM_On_Override=0x30,
                PGM_Off_Override=0x31,
                PGM_On=0x32,
                PGM_Off=0x33,
                Reload_RAM=0x80,
                Bus_Scan=0x85,
                Future_Use=0x90,
            ),
            "argument"
            / Enum(
                Int8ub, One_Beep=0x04, Fail_Beep=0x08, Beep_Twice=0x0C, Accept_Beep=0x10
            ),
            "instant" / Default(Flag, False),
            "_not_used1" / Padding(28),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

PerformActionResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x4, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "_not_used0" / Padding(1),
            "action"
            / Enum(
                Int8ub,
                Stay_Arm=0x01,
                Stay_Arm1=0x02,
                Sleep_Arm=0x03,
                Full_Arm=0x04,
                Disarm=0x05,
                Stay_Arm_StayD=0x06,
                Sleep_Arm_StayD=0x07,
                Disarm_Both_Disable_StayD=0x08,
                Bypass=0x10,
                Beep=0x10,
                PGM_On_Override=0x30,
                PGM_Off_Overrite=0x31,
                PGM_On=0x32,
                PGM_Of=0x33,
                Reload_RAM=0x80,
                Bus_Scan=0x85,
                Future_Use=0x90,
            ),
            "_not_used1" / Padding(33),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

# EEPROM Structures
DefinitionsParserMap = {
    "zone": Struct(
        "definition"
        / Enum(
            Int8ul,
            disabled=0,
            delay_1=1,
            delay_2=2,
            entry_delay_1_full=3,
            entry_delay_2_full=4,
            follow=5,
            follow_sleep_full=6,
            follow_full=7,
            instant=8,
            instant_sleep_full=9,
            instant_full=10,
            instant_fire=11,
            delayed_fire=12,
            instant_fire_silent=13,
            delayed_fire_silent=14,
            buzzer_24h=15,
            burglary_24h=16,
            hold_up_24h=17,
            gas_24hr=18,
            heat_24h=19,
            water_24h=20,
            freeze_24h=21,
            panic_24h=22,
            follow_no_pre_alarm=23,
            instant_no_pre_alarm=24,
            keyswitch_maintain=25,
            keyswitch_momentary=26,
            instant_no_pre_alarm_stay=33,
            instant_no_pre_alarm_sleep=34,
            entry_delay_1_stay_full_instant=35,
            entry_delay_1_full_instant=36,
        ),
        "partition" / Int8ul,
        "options"
        / BitStruct(
            "auto_zone_shutdown" / Flag,
            "bypassable" / Flag,
            "rf_supervision" / Flag,
            "alarm_type"
            / Enum(
                BitsInteger(2),
                audible_alarm_steady=0,
                silent_alarm=1,
                audible_alarm_pulse=2,
                report_only=3,
            ),
            "intellizone" / Flag,
            "delay_alarm_transmission" / Flag,
            "force_arming" / Flag,
        ),
    )
}

SendPanicAction = Struct(  # Supported on firmware versions 7.15+
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "packet_length" / Const(37, Int8ub),
            "unknown0" / Const(0x1A, Int8ub),
            "partition" / Int8ub,
            "panic_type" / Enum(Int8ub, emergency=0, medical=1, fire=2,),  # wild guess
            "user_id" / Int8ub,
            "_not_used" / Padding(26),
            "unknown1" / Const(0x06, Int8ub),
            "_not_used" / Padding(1),
            "user_id" / Int16ul,
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)



import logging
import re
import string

logger = logging.getLogger("PAI").getChild(__name__)

re_magick_placeholder = re.compile("@(?P<type>[a-z]+)(:?#(?P<source>[a-z0-9_]+))?")


class EventMessageFormatter(string.Formatter):
    @staticmethod
    def _hasattr(event, key):
        if key in event.additional_data:
            return True
        return hasattr(event, key)

    @staticmethod
    def _getattr(event, key):
        if key in event.additional_data:
            return event.additional_data[key]
        return getattr(event, key)

    def get_value(self, key, args, kwargs):
        event = args[0]

        if isinstance(key, int) or isinstance(key, float):
            return key
        elif key.startswith("@"):  # pure magic is happening here
            label_provider = event.label_provider
            m = re_magick_placeholder.match(key)
            if m:
                element_type = m.group("type")
                source = m.group("source") or (
                    element_type if self._hasattr(event, element_type) else "minor"
                )
                key = self._getattr(event, source)

                return label_provider(element_type, key)
            else:
                logger.error('Magic placeholder "{}" has wrong format'.format(key))
                return "{error}"

        r = self._getattr(event, key.lower())
        if key[0].isupper():
            r = r.title()

        return r



# -*- coding: utf-8 -*-

import logging

from paradox.config import config as cfg
from paradox.event import (ChangeEvent, Event, EventLevel, LiveEvent,
                           Notification)
from paradox.interfaces.text.core import AbstractTextInterface
from paradox.lib import ps
from paradox.lib.event_filter import EventTagFilter

logger = logging.getLogger("PAI").getChild(__name__)


class DummyInterface(AbstractTextInterface):
    """Interface Class using Dummy"""

    def __init__(self, alarm):
        min_level = EventLevel.from_name(cfg.DUMMY_MIN_EVENT_LEVEL)
        event_filter = EventTagFilter(
            queries=cfg.DUMMY_EVENT_FILTERS, min_level=min_level
        )
        super().__init__(alarm, event_filter=event_filter)

    def handle_notify(self, notification: Notification):
        if self.notification_filter(notification):
            logger.log(
                notification.level,
                "sender: %s, message: %s" % (notification.sender, notification.message),
            )

    def handle_panel_event(self, event: Event):
        if self.event_filter.match(event):
            level = event.level
            # logger.log(level.value, event)
            if isinstance(event, LiveEvent):
                logger.log(level.value, "LiveEvent message: %s" % event.message)
            elif isinstance(event, ChangeEvent):
                logger.log(level.value, "ChangeEvent message: %s" % event.message)
            else:
                logger.log(
                    level.value,
                    "%s message: %s" % (event.__class__.__name__, event.message),
                )



# -*- coding: utf-8 -*-

# IP Interface

import asyncio
import logging

from paradox.config import config as cfg
from paradox.interfaces import Interface

from ...lib.handlers import PersistentHandler
from .client_connection import ClientConnection

logger = logging.getLogger("PAI").getChild(__name__)


class IPInterface(Interface):
    def __init__(self, alarm):
        super().__init__(alarm)
        self.key = cfg.IP_INTERFACE_PASSWORD
        self.addr = cfg.IP_INTERFACE_BIND_ADDRESS
        self.port = cfg.IP_INTERFACE_BIND_PORT
        self.server = None
        self.started = False
        self.client_nr = 0

    # def on_connection_lost(self):
    #     logger.error('Connection with client was lost')

    def stop(self):
        if self.server is not None:
            self.server.close()
            self.server = None
        self.started = False

    def start(self):
        logger.info("Starting %s Interface", self.name)
        self.started = True

        if not self.alarm:
            logger.info("No alarm set")
            return

        asyncio.get_event_loop().create_task(self.run())

    async def run(self):
        try:
            self.server = await asyncio.start_server(
                self.handle_client, self.addr, self.port, loop=self.alarm.work_loop
            )
            logger.info(
                "IP Interface: serving on {}".format(
                    self.server.sockets[0].getsockname()
                )
            )
            logger.info("IP Interface started")
        except Exception as e:
            logger.error("Failed to start IP Interface {}".format(e))

    async def handle_client(self, reader, writer):
        """
        Handle message from the remote client.

        :param reader: Socket read stream from the client
        :param writer: Socket write stream to the client
        :return: None
        """

        self.client_nr = (self.client_nr + 1) % 256
        logger.info("Client %d connected" % self.client_nr)

        connection = ClientConnection(reader, writer, self.alarm, self.key)
        handler_name = "%s_%d" % (self.name, self.client_nr)
        self.alarm.connection.register_raw_handler(
            PersistentHandler(connection.handle_panel_raw_message, name=handler_name)
        )

        await self.alarm.pause()

        try:
            await connection.handle()
        except:
            logger.exception("Client %d connection raised exception" % self.client_nr)
        finally:
            self.alarm.connection.deregister_raw_handler(handler_name)

            asyncio.get_event_loop().create_task(self.alarm.resume())
            logger.info("Client %d disconnected" % self.client_nr)



# -*- coding: utf-8 -*-

import binascii
import inspect
import logging
import typing

from construct import ChecksumError, Construct, Container, MappingError

from paradox.config import config as cfg
from paradox.exceptions import AuthenticationFailed, StatusRequestException

from ..panel import Panel as PanelBase
from . import parsers
from .event import event_map
from .property import property_map

logger = logging.getLogger("PAI").getChild(__name__)

ZONE_ACTIONS = dict(
    bypass={"flags": {"bypassed": True}, "operation": "set",},
    clear_bypass={"flags": {"bypassed": True}, "operation": "clear",},
    clear_alarm_memory={"flags": {"generated_alarm": True}, "operation": "clear",},
)


class Panel_EVOBase(PanelBase):
    event_map = event_map
    property_map = property_map
    max_eeprom_response_data_length = 64
    status_request_addresses = parsers.RAMDataParserMap.keys()

    def __init__(
        self, core, start_communication_response, variable_message_length=True
    ):
        super(Panel_EVOBase, self).__init__(core, variable_message_length)

        self._populate_settings(start_communication_response)

    def _populate_settings(self, start_communication_response):
        raw_data = (
                start_communication_response.fields.data
                + start_communication_response.checksum
        )
        parsed = parsers.InitializeCommunication.parse(raw_data)
        if cfg.LOGGING_DUMP_MESSAGES:
            logger.debug("Second parse of InitializeCommunication response: %s", parsed)
        self.settings = parsed.fields.value

    def get_message(self, name: str) -> Construct:
        try:
            clsmembers = dict(inspect.getmembers(parsers))
            if name in clsmembers:
                return clsmembers[name]
        except ResourceWarning:
            pass

        return super(Panel_EVOBase, self).get_message(name)

    async def dump_memory(self, file, memory_type):
        """
        Dumps EEPROM and RAM memory to files
        :return:
        """
        if memory_type == "ram":
            await self.dump_memory_to_file(file, range(0, 59), True)
        elif memory_type == "eeprom":
            await self.dump_memory_to_file(file, range(0, 0xFFFF, 64))
        else:
            raise AttributeError(f"Unknown memory type: {memory_type}")

    async def dump_memory_to_file(self, file, range_, ram=False):
        mem_type = "RAM" if ram else "EEPROM"
        logger.info("Dump " + mem_type)

        packet_length = 64  # 64 is max
        for address in range_:
            args = dict(
                address=address, length=packet_length, control=dict(ram_access=ram)
            )
            logger.info("Dumping %s: address %d" % (mem_type, address))
            reply = await self.core.send_wait(
                parsers.ReadEEPROM,
                args,
                reply_expected=lambda m: m.fields.value.po.command == 0x05
                and m.fields.value.address == address,
            )

            if reply is None:
                logger.error("Could not read %s: address %d" % (mem_type, address))
                return

            data = reply.fields.value.data

            file.write(data)

    def parse_message(
        self, message: bytes, direction="topanel"
    ) -> typing.Optional[Container]:
        try:
            if message is None or len(message) == 0:
                return None

            parent_parsed = super(Panel_EVOBase, self).parse_message(message, direction)
            if parent_parsed:
                return parent_parsed

            if direction == "topanel":
                if message[0] == 0x70:
                    return parsers.CloseConnection.parse(message)
                elif message[0] == 0x00:
                    return parsers.InitializeCommunication.parse(message)
                elif message[0] == 0x30:
                    return parsers.SetTimeDate.parse(message)
                elif message[0] == 0x40:
                    return parsers.PerformPartitionAction.parse(message)
                elif message[0] == 0xD0:
                    return parsers.PerformZoneAction.parse(message)
            else:
                if message[0] >> 4 == 0x7:
                    return parsers.ErrorMessage.parse(message)
                elif message[0] >> 4 == 0x1:
                    return parsers.LoginConfirmationResponse.parse(message)
                elif message[0] >> 4 == 0x03:
                    return parsers.SetTimeDateResponse.parse(message)
                elif message[0] >> 4 == 4:  # Used for partitions and PGMs
                    return parsers.PerformActionResponse.parse(message)
                elif message[0] >> 4 == 0xD:
                    return parsers.PerformZoneActionResponse.parse(message)
                # elif message[0] == 0x50 and message[2] == 0x80:
                #     return PanelStatus.parse(message)
                # elif message[0] == 0x50 and message[2] < 0x80:
                #     return ReadEEPROM.parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] == 0x80:
                #     return PanelStatusResponse[message[3]].parse(message)
                # elif message[0] >> 4 == 0x05 and message[2] < 0x80:
                elif message[0] >> 4 == 0x05:
                    return parsers.ReadEEPROMResponse.parse(message)
                # elif message[0] == 0x60 and message[2] < 0x80:
                #     return WriteEEPROM.parse(message)
                # elif message[0] >> 4 == 0x06 and message[2] < 0x80:
                #     return WriteEEPROMResponse.parse(message)
                elif message[0] >> 4 == 0x0E:
                    if message[1] == 0xFF:
                        return parsers.LiveEvent.parse(message)
                    else:
                        return parsers.RequestedEvent.parse(message)

        except ChecksumError as e:
            logger.error(
                "ChecksumError %s, message: %s" % (str(e), binascii.hexlify(message))
            )
        except:
            logger.exception(
                "Exception parsing message: %s" % (binascii.hexlify(message))
            )

        return None

    async def initialize_communication(self, password) -> bool:
        encoded_password = self.encode_password(password)

        self.settings.pc_password = encoded_password
        payload = parsers.InitializeCommunication.build(
            dict(fields=dict(value=self.settings))
        )

        logger.info("Installer login")
        reply = await self.core.send_wait(message=payload, reply_expected=[0x1, 0x0])

        if reply is None:
            logger.error("Installer login failed")
            return False

        if reply.fields.value.po.command == 0x0:
            logger.error(
                "Authentication Failed. Wrong PASSWORD. Make sure you use correct PC Password. In Babyware: "
                "Right click on your panel -> Properties -> PC Communication (Babyware) -> PC Communication "
                "(Babyware) Tab."
            )
            raise AuthenticationFailed("Wrong PASSWORD")
        else:  # command == 0x1
            if reply.fields.value.po.status.Winload_connected:
                logger.info("Authentication Success")
                return True
            else:
                logger.error("Authentication Failed")
                return False

    @staticmethod
    def _request_status_reply_check(message: Container, address: int):
        mvars = message.fields.value

        if (
            mvars.po.command == 0x5
            and mvars.control.ram_access is True
            and mvars.bus_address == 0x00  # panel
            and mvars.address == address
        ):
            return True

        return False

    async def request_status(self, i: int) -> typing.Optional[Container]:
        args = dict(address=i, length=64, control=dict(ram_access=True))
        reply = await self.core.send_wait(
            parsers.ReadEEPROM,
            args,
            reply_expected=lambda m: self._request_status_reply_check(
                m, args["address"]
            ),
        )
        if reply is not None:
            logger.debug("Received status response: %d" % i)
            return self.handle_status(reply, parsers.RAMDataParserMap)
        else:
            raise StatusRequestException("No reply to status request: %d" % i)

    async def control_partitions(self, partitions: list, command: str) -> bool:
        """
        Control Partitions
        :param list partitions: a list of partitions
        :param str command: textual command
        :return: True if we have at least one success
        """
        args = dict(partitions=dict((i, command) for i in partitions))

        try:
            reply = await self.core.send_wait(
                parsers.PerformPartitionAction, args, reply_expected=0x04
            )
        except MappingError:
            logger.error('Partition command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Partition command: "%s" succeeded' % command)
        else:
            logger.info('Partition command: "%s" failed' % command)
        return reply is not None

    async def control_zones(self, zones: list, command: str) -> bool:
        """
        Control zones
        :param list zones: a list of zones
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in ZONE_ACTIONS:
            return False

        args = ZONE_ACTIONS[command].copy()
        args["zones"] = zones

        try:
            reply = await self.core.send_wait(
                parsers.PerformZoneAction, args, reply_expected=0xD
            )
        except MappingError:
            logger.error('Zone command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Zone command: "%s" succeeded' % command)
        else:
            logger.info('Zone command: "%s" failed' % command)
        return reply is not None

    async def control_outputs(self, outputs, command) -> bool:
        """
        Control PGM
        :param list outputs: a list of pgms
        :param str command: textual command
        :return: True if we have at least one success
        """

        args = {"pgms": outputs, "command": command}

        try:
            reply = await self.core.send_wait(
                parsers.PerformPGMAction, args, reply_expected=0x04
            )
        except MappingError:
            logger.error('PGM command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('PGM command: "%s" succeeded' % command)
        else:
            logger.info('PGM command: "%s" failed' % command)
        return reply is not None

    async def control_doors(self, doors, command) -> bool:
        """
        Control Doors
        :param list doors: a list of doors
        :param str command: textual command
        :return: True if we have at least one success
        """

        args = {"doors": doors, "command": command}

        try:
            reply = await self.core.send_wait(
                parsers.PerformDoorAction, args, reply_expected=0x04
            )
        except MappingError:
            logger.error('Door command: "%s" is not supported' % command)
            return False

        if reply:
            logger.info('Door command: "%s" succeeded' % command)
        else:
            logger.info('Door command: "%s" failed' % command)
        return reply is not None

    async def send_panic(self, partitions, panic_type, user_id):
        accepted = False

        args = {"partitions": partitions, "panic_type": panic_type, "user_id": user_id}

        reply = await self.core.send_wait(
            parsers.SendPanicAction, args, reply_expected=0x04
        )

        if reply is not None:
            accepted = True

        return accepted



from .models.evo48 import Panel_EVO48
from .models.evo96 import Panel_EVO96
from .models.evo192 import Panel_EVO192
from .models.evohd import Panel_EVOHD

__all__ = ["Panel_EVO48", "Panel_EVO96", "Panel_EVO192", "Panel_EVOHD"]



# -*- coding: utf-8 -*-
import asyncio
import json
import logging
import time

from pushbullet import Pushbullet
from ws4py.client import WebSocketBaseClient
from ws4py.manager import WebSocketManager

from paradox.config import config as cfg
from paradox.event import EventLevel, Notification
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface
from paradox.lib import ps

# Pushbullet interface.
# Only exposes critical status changes and accepts commands

logger = logging.getLogger("PAI").getChild(__name__)


class PushBulletWSClient(WebSocketBaseClient):
    name = "pushbullet"

    def __init__(self, interface, url):
        """ Initializes the PB WS Client"""
        super().__init__(url)

        self.pb = Pushbullet(cfg.PUSHBULLET_KEY)
        self.manager = WebSocketManager()
        self.interface = interface

        self.device = None
        for i, device in enumerate(self.pb.devices):
            if device.nickname == cfg.PUSHBULLET_DEVICE:
                logger.debug("Device found")
                self.device = device
                break
        else:
            logger.exception("Device not found. Creating 'pai' device")
            self.device = self.pb.new_device(nickname="pai", icon="system")

    def stop(self):
        self.terminate()
        self.manager.stop()

    def handshake_ok(self):
        """ Callback trigger when connection succeeded"""
        logger.info("Handshake OK")
        self.manager.add(self)
        self.manager.start()
        for chat in self.pb.chats:
            logger.debug("Associated contacts: {}".format(chat))

        # Receiving pending messages
        self.received_message(json.dumps({"type": "tickle", "subtype": "push"}))

        self.send_message("Active")

    def received_message(self, message):
        """ Handle Pushbullet message. It should be a command """
        logger.debug("Received Message {}".format(message))

        try:
            message = json.loads(str(message))
        except:
            logger.exception("Unable to parse message")
            return

        if message["type"] == "tickle" and message["subtype"] == "push":
            now = time.time()
            pushes = self.pb.get_pushes(
                modified_after=int(now) - 20, limit=1, filter_inactive=True
            )
            for p in pushes:

                # Ignore messages send by us
                if p.get("direction") == "self" and p.get("title") == "pai":
                    # logger.debug('Ignoring message sent')
                    continue

                if p.get("direction") == "outgoing" or p.get("dismissed"):
                    # logger.debug('Ignoring outgoing dismissed')
                    continue

                if (
                    p.get("sender_email_normalized") in cfg.PUSHBULLET_CONTACTS
                    or p.get("direction") == "self"
                ):
                    future = asyncio.run_coroutine_threadsafe(
                        self.interface.handle_command(p.get("body")),
                        self.interface.alarm.work_loop,
                    )
                    ret = future.result(10)

                    m = "PB {}: {}".format(p.get("sender_email_normalized"), ret)
                    logger.info(m)
                else:
                    m = "PB {} (UNK): {}".format(
                        p.get("sender_email_normalized"), p.get("body")
                    )
                    logger.warning(m)

                self.send_message(m)
                ps.sendNotification(
                    Notification(sender=self.name, message=m, level=EventLevel.INFO)
                )

    def unhandled_error(self, error):
        logger.error("{}".format(error))

        try:
            self.terminate()
        except:
            logger.exception("Closing Pushbullet WS")

        self.close()

    def send_message(self, msg, dstchat=None):
        if dstchat is None:
            dstchat = self.pb.chats

        if not isinstance(dstchat, list):
            dstchat = [dstchat]
        # Push to self
        self.device.push_note(cfg.PUSHBULLET_DEVICE, msg)

        for chat in dstchat:
            if chat.email in cfg.PUSHBULLET_CONTACTS:
                try:
                    self.pb.push_note(cfg.PUSHBULLET_DEVICE, msg, chat=chat)
                except:
                    logger.exception("Sending message")
                    time.sleep(5)


class PushbulletTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Pushbullet"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.PUSHBULLET_EVENT_FILTERS,
            cfg.PUSHBULLET_ALLOW_EVENTS,
            cfg.PUSHBULLET_IGNORE_EVENTS,
            cfg.PUSHBULLET_MIN_EVENT_LEVEL,
        )
        self.name = PushBulletWSClient.name
        self.pb_ws = None

    def _run(self):
        super(PushbulletTextInterface, self)._run()
        try:
            self.pb_ws = PushBulletWSClient(
                self,
                "wss://stream.pushbullet.com/websocket/{}".format(cfg.PUSHBULLET_KEY),
            )
            self.pb_ws.connect()
        except:
            logger.exception("Could not connect to Pushbullet service")

        logger.info("Pushbullet Interface Started")

    def stop(self):
        """ Stops the Pushbullet interface"""
        super().stop()
        if self.pb_ws is not None:
            self.pb_ws.stop()

    def send_message(self, message: str, level: EventLevel):
        if self.pb_ws is not None:
            self.pb_ws.send_message(message)



#!/usr/bin/env python3

__author__ = "João Paulo Barraca, Jevgeni Kiski"
__copyright__ = "Copyright 2018-2019, João Paulo Barraca"
__credits__ = ["Tihomir Heidelberg", "Louis Rossouw"]
__license__ = "EPL"
__maintainer__ = "João Paulo Barraca"
__email__ = "jpbarraca@gmail.com"
__status__ = "Beta"

import argparse
import sys

if sys.version_info < (3, 6,):
    print(
        "You are using Python %s.%s, but PAI requires at least Python 3.6"
        % (sys.version_info[0], sys.version_info[1])
    )
    sys.exit(-1)


def main():
    from paradox.main import main

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        help="specify path to an alternative configuration file",
    )

    args = parser.parse_args()

    main(args)


if __name__ == "__main__":
    from paradox.lib import help

    try:
        main()
    except ImportError as error:
        help.import_error_help(error)



import asyncio
import binascii
import logging
import os
from enum import Enum, auto
from typing import Awaitable, Union

from construct import Container

from paradox.config import config as cfg
from paradox.connections.ip.parsers import (IPMessageCommand, IPMessageRequest,
                                            IPMessageResponse, IPMessageType,
                                            IPPayloadConnectResponse)

logger = logging.getLogger("PAI").getChild(__name__)


class Status(Enum):
    CONNECTING = (auto(),)
    CONNECTED = (auto(),)
    CLOSING_CONNECTION = auto()


class ClientConnection:
    def __init__(self, reader, writer, alarm, key):
        self.status = Status.CONNECTING
        if isinstance(key, str):
            key = key.encode()
        self.client_writer = writer
        self.client_reader = reader
        self.alarm = alarm
        self.interface_password = key
        self.connection_key = key
        self.flags = Container(encrypt=True)

    async def handle_panel_raw_message(self, raw: bytes):
        """
        Handle message from panel, which must be sent to the client
        """

        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.serial_passthrough_response,
                command=IPMessageCommand.passthrough,
                flags=Container(
                    keep_alive=True,
                    live_events=True,
                    neware=True,
                    upload_download=True,
                    encrypt=self.flags.encrypt,
                ),
                sb=3,
                wt=0,
            ),
            payload=raw,
        )

        await self._send_to_client(out_message_container)

    async def handle(self):
        while True:
            try:
                data = await self._client_read()
            except asyncio.TimeoutError:
                logger.info("Timeout. Client may have disconnected")
                break
            except:
                logger.info("Client disconnected")
                break

            if not data:
                if not self.alarm.connection.connected:
                    break
                continue

            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug("APP -> IPI (raw) {}".format(binascii.hexlify(data)))

            in_message = IPMessageRequest.parse(data, password=self.connection_key)

            if cfg.LOGGING_DUMP_MESSAGES:
                logger.debug("APP -> IPI (message) {}".format(in_message))

            if cfg.LOGGING_DUMP_PACKETS:
                logger.debug(
                    "APP -> IPI (payload) {}".format(
                        binascii.hexlify(in_message.payload)
                    )
                )

            if in_message.header.message_type == IPMessageType.ip_request:
                if in_message.header.command == IPMessageCommand.connect:
                    await self._handle_ip_authentication(in_message)
                elif in_message.header.command == IPMessageCommand.keep_alive:
                    await self._handle_keep_alive(in_message)
                elif (
                    in_message.header.command
                    == IPMessageCommand.upload_download_connection
                ):
                    await self._handle_upload_download_connection(in_message)
                elif (
                    in_message.header.command
                    == IPMessageCommand.upload_download_disconnection
                ):
                    await self._handle_upload_download_disconnection(in_message)
                elif in_message.header.command == IPMessageCommand.toggle_keep_alive:
                    await self._handle_toggle_keep_alive(in_message)

                # FB - Multicommand
                # F5 - no idea
                # elif message.header.command in ["F5", "FB"]:  # Proxy Insite Gold communication
                #     TODO: Implement
                #     if not isinstance(self.alarm.connection, IPConnection):
                #         logger.error(f"Only IP Connection supports '{message.header.command}' command")
                #         break
                #
                #     proxy_message = Container(
                #         header=message.header,
                #         payload=in_payload
                #     )
                #
                #     async with self.alarm.request_lock, self.alarm.busy:
                #         self.alarm.connection.write_with_header(IPMessageRequest(proxy_message))

                else:
                    logger.warning(
                        f"Unknown ip_request: raw: {binascii.hexlify(data)}, message: {in_message}"
                    )
                    continue
            elif (
                in_message.header.message_type
                == IPMessageType.serial_passthrough_request
            ):
                if in_message.header.command == IPMessageCommand.passthrough:
                    await self._handle_passthrough(in_message)

                else:
                    logger.warning(
                        f"Unknown serial_passthrough_request: raw: {binascii.hexlify(data)}, message: {in_message}"
                    )

            if self.status == Status.CLOSING_CONNECTION:
                break

    async def _handle_passthrough(self, in_message):
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.message_type = (
            IPMessageType.serial_passthrough_response
        )
        out_message_container.header.flags.upload_download = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.header.flags.keep_alive = True
        if (
            in_message.payload[0] == 0x70 and in_message.payload[2] == 0x05
        ):  # Close connection
            out_message_container.payload = self.alarm.panel.get_message(
                "CloseConnection"
            ).build({})
            self.status = Status.CLOSING_CONNECTION
        else:
            async with self.alarm.request_lock, self.alarm.busy:
                self.alarm.connection.write(in_message.payload)
        if in_message.payload[0] == 0x00:  # Just a status update
            self.status = Status.CONNECTED

    async def _handle_ip_authentication(self, in_message):
        logger.info("Authenticating to IP interface")
        password = in_message.payload
        if password != self.interface_password:
            await self._handle_failed_login(in_message)
        else:
            await self._handle_successful_login(in_message)

    async def _send_to_client(self, container):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug(
                "IPI -> APP (payload) {}".format(binascii.hexlify(container.payload))
            )
        raw = IPMessageResponse.build(container, password=self.connection_key)
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("IPI -> APP (raw) {}".format(binascii.hexlify(raw)))
        self.client_writer.write(raw)
        await self.client_writer.drain()

    async def _handle_toggle_keep_alive(self, in_message):
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.upload_download = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.header.flags.keep_alive = True
        out_message_container.payload = b"\x01"

        await self._send_to_client(out_message_container)

    async def _handle_upload_download_disconnection(self, in_message):
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.payload = (
            b"\x01" if self.status == Status.CLOSING_CONNECTION else b"\x00"
        )

        await self._send_to_client(out_message_container)

    async def _handle_upload_download_connection(self, in_message):
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.upload_download = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.payload = binascii.unhexlify(
            "0100000000000000000000000000000000"
        )

        await self._send_to_client(out_message_container)

    async def _handle_keep_alive(self, in_message):
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.payload = b"\x00"

        await self._send_to_client(out_message_container)

    async def _handle_successful_login(self, in_message):
        logger.info("Authentication Success")
        # Generate a new key
        next_connection_key = binascii.hexlify(os.urandom(8)).upper()

        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.neware = True
        out_message_container.header.flags.live_events = True
        out_message_container.payload = IPPayloadConnectResponse.build(
            dict(
                key=next_connection_key,
                login_status="success",
                hardware_version=32,
                ip_firmware_major=1,
                ip_firmware_minor=32,
                ip_module_serial=b"\x71\x23\x45\x67",  # 0x71 = IP150
            )
        )

        await self._send_to_client(out_message_container)

        self.connection_key = next_connection_key

    async def _handle_failed_login(self, in_message):
        logger.warning("Authentication Error: Wrong password")
        out_message_container = Container(
            header=Container(
                message_type=IPMessageType.ip_response,
                command=in_message.header.command,
                flags=Container(encrypt=in_message.header.flags.encrypt),
                sb=3,
                wt=0,
            )
        )
        out_message_container.header.flags.encrypt = False
        out_message_container.header.flags.installer_mode = True
        out_message_container.header.flags.neware = True
        out_message_container.payload = IPPayloadConnectResponse.build(
            dict(
                key=b"\00" * 16,
                login_status="invalid_password",
                hardware_version=0,
                ip_firmware_major=0,
                ip_firmware_minor=0,
                ip_module_serial=b"\x71\x00\x00\x00",
            )
        )
        self.status = Status.CLOSING_CONNECTION

        await self._send_to_client(out_message_container)

    async def _client_read(self) -> bytes:
        return await asyncio.wait_for(
            self.client_reader.read(1000), cfg.KEEP_ALIVE_INTERVAL * 1.5
        )



import logging
import os

import requests

from paradox.config import config as cfg
from paradox.event import EventLevel
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface

logger = logging.getLogger("PAI").getChild(__name__)


class HomeAssistantNotificationsTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Home Assistant"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.HOMEASSISTANT_NOTIFICATIONS_EVENT_FILTERS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_ALLOW_EVENTS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_IGNORE_EVENTS,
            cfg.HOMEASSISTANT_NOTIFICATIONS_MIN_EVENT_LEVEL,
        )

        self.token = os.environ.get("SUPERVISOR_TOKEN")
        if not self.token:
            logger.error(
                f'"SUPERVISOR_TOKEN" environment variable must be set to use {__class__.__name__}'
            )

    def send_message(self, message: str, level: EventLevel):
        if not self.token:
            logger.warning(
                'Unable to send a notification to Home Assistant. "SUPERVISOR_TOKEN" environment variable is not set'
            )
            return

        notifier_name = cfg.HOMEASSISTANT_NOTIFICATIONS_NOTIFIER_NAME
        url = f"http://supervisor/core/api/services/notify/{notifier_name}"

        payload = {"message": message, "title": "Paradox", "data": {"level": level}}

        headers = {"Authorization": f"Bearer {self.token}"}

        res = requests.post(url, json=payload, headers=headers)
        if res.status_code == 200:
            logger.debug(f"Notification sent: {message}, level={level}")
        else:
            logger.error(
                f"Failed to send notification: code={res.status_code}, text: {res.text}"
            )



from time import time

from paradox.event import EventLevel


def _toggle_provider(old):
    return not old


def _timestamp_provider(old):
    return int(time())


event_opt1 = {
    1: dict(type="output", id=1),
    2: dict(type="output", id=2),
    3: dict(type="output", id=3),
    4: dict(type="output", id=4),
    5: dict(type="output", id=5),
    6: dict(type="output", id=6),
    7: dict(type="output", id=7),
    8: dict(type="output", id=8),
    9: dict(type="output", id=9),
    10: dict(type="output", id=10),
    11: dict(type="output", id=11),
    12: dict(type="output", id=12),
    13: dict(type="output", id=13),
    14: dict(type="output", id=14),
    15: dict(type="output", id=15),
    16: dict(type="output", id=16),
    17: dict(type="repeater", id=1),
    18: dict(type="repeater", id=2),
    19: dict(type="keypad", id=1),
    20: dict(type="keypad", id=2),
    21: dict(type="keypad", id=3),
    22: dict(type="keypad", id=4),
    27: dict(type="siren", id=1),
    28: dict(type="siren", id=2),
    29: dict(type="siren", id=3),
    30: dict(type="siren", id=4),
    99: dict(),
}

event_map = {
    0: dict(
        level=EventLevel.DEBUG,
        change=dict(open=False),
        type="zone",
        message="Zone {label} OK",
    ),
    1: dict(
        level=EventLevel.DEBUG,
        change=dict(open=True),
        type="zone",
        message="Zone {label} open",
    ),
    2: dict(
        level=EventLevel.CRITICAL,
        type="partition",
        sub={
            0: dict(message="N/A"),
            1: dict(message="N/A"),
            2: dict(
                change=dict(silent_alarm=True, alarm=True),
                tags=["alarm", "silent_alarm"],
                message="Partition {label} silent alarm active",
            ),
            3: dict(
                change=dict(audible_alarm=True, alarm=True),
                tags=["alarm", "buzzer_alarm"],
                message="Partition {label} buzzer alarm active",
            ),
            4: dict(
                change=dict(audible_alarm=True, alarm=True),
                tags=["alarm", "steady_alarm"],
                message="Partition {label} steady alarm active",
            ),
            5: dict(
                change=dict(audible_alarm=True, alarm=True),
                tags=["alarm", "pulse_alarm"],
                message="Partition {label} pulse alarm active",
            ),
            6: dict(
                change=dict(strobe_alarm=True, alarm=True),
                tags=["alarm", "strobe"],
                message="Partition {label} strobe active",
            ),
            7: dict(
                level=EventLevel.INFO,
                change=dict(alarm=False),
                message="Partition {label} alarm stopped",
            ),
            8: dict(
                level=EventLevel.DEBUG,
                change=dict(squawk=True),
                id=1,
                message="Squawk ON",
            ),
            9: dict(
                level=EventLevel.DEBUG,
                change=dict(squawk=False),
                id=1,
                message="Squawk OFF",
            ),
            10: dict(level=EventLevel.DEBUG, message="Ground Start"),
            11: dict(
                level=EventLevel.INFO,
                change=dict(arm=False, arm_sleep=False, arm_stay=False, alarm=False),
                tags=["disarm"],
                message="Partition {label} disarmed",
            ),
            12: dict(
                level=EventLevel.INFO,
                change=dict(arm=True),
                tags=["arm"],
                message="Partition {label} armed",
            ),
            13: dict(
                level=EventLevel.INFO,
                change=dict(entry_delay=True),
                message="Partition {label} in entry delay",
            ),
            14: dict(
                level=EventLevel.INFO,
                change=dict(exit_delay=True),
                message="Partition {label} in exit delay",
            ),
            15: dict(level=EventLevel.DEBUG, message="Pre-alarm delay"),
            16: dict(level=EventLevel.DEBUG, message="Report confirmation"),
            99: dict(level=EventLevel.DEBUG, message="Any partition status event"),
        },
    ),
    3: dict(
        level=EventLevel.DEBUG,
        type="system",
        sub={
            0: dict(change=dict(bell_activated=False), message="Bell OFF"),
            1: dict(change=dict(bell_activated=True), message="Bell ON"),
            2: dict(message="Bell squawk arm"),
            3: dict(message="Bell squawk disarm"),
            99: dict(message="Any bell status event"),
        },
    ),
    # 5: 'Non-Reportable Event',
    6: dict(
        level=EventLevel.DEBUG,
        message="Non-reportable event",
        sub={
            0: dict(
                level=EventLevel.CRITICAL,
                type="system",
                tags=["trouble"],
                message="Telephone line trouble",
            ),
            1: dict(type="system", message="[ENTER]/[CLEAR]/[POWER] key was pressed"),
            2: dict(message="N/A"),
            3: dict(
                level=EventLevel.INFO,
                change=dict(arm_stay=True, arm=True),
                type="partition",
                tags=["arm"],
                message="Arm in stay mode",
            ),
            4: dict(
                level=EventLevel.INFO,
                change=dict(arm_sleep=True, arm=True),
                type="partition",
                tags=["arm"],
                message="Arm in sleep mode",
            ),
            5: dict(
                level=EventLevel.INFO,
                change=dict(arm_force=True, arm=True),
                type="partition",
                tags=["arm"],
                message="Arm in force mode",
            ),
            6: dict(
                level=EventLevel.INFO,
                change=dict(arm_stay=True, arm=True),
                type="partition",
                tags=["arm"],
                message="Full arm when armed in stay mode",
            ),
            7: dict(
                level=EventLevel.CRITICAL,
                tags=["trouble"],
                change=dict(computer_fail_to_communicate_trouble=True),
                type="system",
                message="PC fail to communicate",
            ),
            8: dict(type="system", message="Utility Key 1 pressed (keys [1] and [2])"),
            9: dict(type="system", message="Utility Key 2 pressed (keys [4] and [5])"),
            10: dict(type="system", message="Utility Key 3 pressed (keys [7] and [8])"),
            11: dict(type="system", message="Utility Key 4 pressed (keys [2] and [3])"),
            12: dict(type="system", message="Utility Key 5 pressed (keys [5] and [6])"),
            13: dict(type="system", message="Utility Key 6 pressed (keys [8] and [9])"),
            14: dict(
                level=EventLevel.CRITICAL,
                type="system",
                tags=["alarm"],
                message="Tamper generated alarm",
            ),
            15: dict(
                level=EventLevel.CRITICAL,
                type="system",
                tags=["supervision", "alarm"],
                message="Supervision loss generated alarm",
            ),
            16: dict(message="N/A"),
            17: dict(message="N/A"),
            18: dict(message="N/A"),
            19: dict(message="N/A"),
            20: dict(
                level=EventLevel.INFO,
                change=dict(arm=True),
                type="partition",
                message="Full arm when armed in sleep mode",
            ),
            21: dict(type="firmware", message="Firmware upgrade"),
            22: dict(message="N/A"),
            23: dict(
                change=dict(stayd_mode_active=True),
                type="partition",
                message="StayD mode activated",
            ),
            24: dict(
                change=dict(stayd_mode_active=False),
                type="partition",
                message="StayD mode deactivated",
            ),
            25: dict(type="system", message="IP Registration status change"),
            26: dict(type="system", message="GPRS Registration status change"),
            27: dict(
                level=EventLevel.INFO, type="partition", message="Armed with trouble(s)"
            ),
            28: dict(
                level=EventLevel.WARN,
                type="system",
                tags=["trouble", "supervision"],
                message="Supervision alert",
            ),
            29: dict(
                level=EventLevel.INFO,
                type="system",
                tags=["trouble", "restore", "supervision"],
                message="Supervision alert restore",
            ),
            30: dict(
                level=EventLevel.WARN,
                type="partition",
                message="Armed with remote with low battery",
            ),
            99: dict(message="Any non-reportable event"),
        },
    ),
    # 7: 'PGM Activation',
    8: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_b=_timestamp_provider),
        message="Button B pressed on remote of {@user}",
    ),
    9: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_c=_timestamp_provider),
        message="Button C pressed on remote of {@user}",
    ),
    10: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_d=_timestamp_provider),
        message="Button D pressed on remote of {@user}",
    ),
    11: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_e=_timestamp_provider),
        message="Button E pressed on remote of {@user}",
    ),
    12: dict(
        level=EventLevel.DEBUG, type="user", message="Cold start wireless zone {label}"
    ),
    13: dict(
        level=EventLevel.DEBUG,
        type="user",
        message="Cold start wireless module {label}",
    ),
    14: dict(
        level=EventLevel.DEBUG,
        type="zone",
        change=dict(bypassed=_toggle_provider),
        message="Bypass programming on zone {label}",
    ),
    15: dict(
        level=EventLevel.DEBUG,
        type="partition",
        message="User code activated output {label}",
    ),
    16: dict(
        level=EventLevel.DEBUG, type="zone", message="Wireless smoke maintenance signal"
    ),
    17: dict(
        level=EventLevel.DEBUG,
        type="zone",
        message="Delay zone {label} alarm transmission",
    ),
    18: dict(
        level=EventLevel.WARN,
        type="zone",
        message="Zone {label} signal strength weak 1",
    ),
    19: dict(
        level=EventLevel.WARN,
        type="zone",
        message="Zone {label} signal strength weak 2",
    ),
    20: dict(
        level=EventLevel.WARN,
        type="zone",
        message="Zone {label} signal strength weak 3",
    ),
    21: dict(
        level=EventLevel.WARN,
        type="zone",
        message="Zone {label} signal strength weak 4",
    ),
    22: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_5=_timestamp_provider),
        message="Button 5 pressed on remote of {@user}",
    ),
    23: dict(
        level=EventLevel.DEBUG,
        type="user",
        change=dict(button_6=_timestamp_provider),
        message="Button 6 pressed on remote of {@user}",
    ),
    24: dict(
        level=EventLevel.WARN,
        type="zone",
        change=dict(fire_delay=True),
        message="Fire delay started",
    ),
    # 25: 'N/A',
    26: dict(
        level=EventLevel.DEBUG,
        type="system",
        message="Software access",
        sub={
            0: dict(message="Non-valid source ID"),
            1: dict(message="WinLoad direct"),
            2: dict(message="WinLoad through IP module"),
            3: dict(message="WinLoad through GSM module"),
            4: dict(message="WinLoad through modem"),
            9: dict(message="IP100 direct"),
            10: dict(message="VDMP3 direct"),
            11: dict(message="Voice through GSM module"),
            12: dict(message="Remote access"),
            13: dict(message="SMS through GSM module"),
            99: dict(message="Any software access"),
        },
    ),
    27: dict(
        level=EventLevel.DEBUG,
        type="bus",
        message="Bus module event",
        sub={
            0: dict(message="A bus module was added"),
            1: dict(message="A bus module was removed"),
            2: dict(message="2-way RF Module Communication Failure"),
            3: dict(message="2-way RF Module Communication Restored"),
        },
    ),
    28: dict(
        level=EventLevel.INFO,
        type="zone",
        message="StayD pass acknowledged for {label}",
    ),
    29: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["arm"],
        message="Arming by user {label}",
    ),
    30: dict(
        level=EventLevel.INFO,
        type="special",
        message="Special arming",
        sub={
            0: dict(tags=["arm"], message="Auto-arming (on time/no movement)"),
            1: dict(message="Late to close"),
            2: dict(tags=["arm"], message="No movement arming"),
            3: dict(tags=["arm"], message="Partial arming"),
            4: dict(tags=["arm"], message="Quick arming"),
            5: dict(tags=["arm"], message="Arming through WinLoad"),
            6: dict(tags=["arm"], message="Arming with keyswitch"),
            99: dict(tags=["arm"], message="Any special arming"),
        },
    ),
    31: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["disarm"],
        message="Disarming by user {label}",
    ),
    32: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["disarm", "after_alarm"],
        message="Disarming after alarm by user {label}",
    ),
    33: dict(
        level=EventLevel.INFO,
        type="user",
        tags=["alarm", "cancel"],
        message="Alarm cancelled by user {label}",
    ),
    34: dict(
        level=EventLevel.WARN,
        type="special",
        sub={
            0: dict(message="Auto-arm cancelled (on time/no movement)"),
            1: dict(
                level=EventLevel.INFO,
                tags=["disarm"],
                message="Disarming through WinLoad",
            ),
            2: dict(
                level=EventLevel.INFO,
                tags=["disarm", "after_alarm"],
                message="Disarming through WinLoad after alarm",
            ),
            3: dict(
                level=EventLevel.INFO,
                tags=["alarm", "cancel"],
                message="Alarm cancelled through WinLoad",
            ),
            4: dict(
                level=EventLevel.INFO,
                tags=["alarm", "cancel"],
                message="Paramedical alarm cancelled",
            ),
            5: dict(
                level=EventLevel.DEBUG, tags=["disarm"], message="Disarm with keyswitch"
            ),
            6: dict(
                level=EventLevel.DEBUG,
                tags=["disarm", "after_alarm"],
                message="Disarm with keyswitch after an alarm",
            ),
            7: dict(
                level=EventLevel.INFO,
                tags=["alarm", "cancel"],
                message="Alarm cancelled with keyswitch",
            ),
            99: dict(
                level=EventLevel.INFO, tags=["disarm"], message="Any special disarming"
            ),
        },
    ),
    35: dict(
        level=EventLevel.INFO,
        change=dict(bypassed=_toggle_provider),
        type="zone",
        tags=["bypass"],
        message="Zone {label} bypass toggled",
    ),
    36: dict(
        level=EventLevel.CRITICAL,
        change=dict(alarm=True),
        type="zone",
        tags=["alarm", "trigger"],
        message="Zone {label} in alarm",
    ),
    37: dict(
        level=EventLevel.CRITICAL,
        change=dict(fire_alarm=True),
        type="zone",
        tags=["alarm", "fire", "trigger"],
        message="Zone {label} in fire alarm",
    ),
    38: dict(
        level=EventLevel.CRITICAL,
        change=dict(alarm=False),
        type="zone",
        tags=["alarm", "restore"],
        message="Zone {label} alarm restore",
    ),
    39: dict(
        level=EventLevel.CRITICAL,
        change=dict(fire_alarm=False),
        type="zone",
        tags=["alarm", "fire", "restore"],
        message="Zone {label} fire alarm restore",
    ),
    40: dict(
        level=EventLevel.CRITICAL,
        type="special",
        tags=["alarm", "trigger"],
        message="Special alarm",
        sub={
            0: dict(message="Panic non-medical emergency"),
            1: dict(message="Panic medical"),
            2: dict(message="Panic fire"),
            3: dict(message="Recent closing"),
            4: dict(message="Global shutdown"),
            5: dict(message="Duress alarm"),
            6: dict(message="Keypad lockout"),
            99: dict(message="Any special alarm event"),
        },
    ),
    41: dict(
        level=EventLevel.INFO,
        change=dict(shutdown=True),
        type="zone",
        message="Zone {label} shutdown",
    ),
    42: dict(
        level=EventLevel.WARN,
        change=dict(zone_tamper_trouble=True),
        tags=["tamper"],
        type="zone",
        message="Zone {label} tampered",
    ),
    43: dict(
        level=EventLevel.INFO,
        change=dict(zone_tamper_trouble=False),
        tags=["tamper", "restore"],
        type="zone",
        message="Zone {label} tamper restore",
    ),
    44: dict(
        level=EventLevel.WARN,
        type="system",
        tags=["trouble"],
        sub={
            0: dict(message="N/A"),
            1: dict(message="AC failure"),
            2: dict(message="Battery failure"),
            3: dict(message="Auxiliary current overload"),
            4: dict(message="Bell current overload"),
            5: dict(message="Bell disconnected"),
            6: dict(change=dict(timer_loss_trouble=True), message="Clock loss"),
            7: dict(change=dict(fire_loop_trouble=True), message="Fire loop trouble"),
            8: dict(message="Fail to communicate to monitoring station telephone #1"),
            9: dict(message="Fail to communicate to monitoring station telephone #2"),
            11: dict(message="Fail to communicate to voice report"),
            12: dict(message="RF jamming"),
            13: dict(message="GSM RF jamming"),
            14: dict(message="GSM no service"),
            15: dict(message="GSM supervision lost"),
            16: dict(message="Fail To Communicate IP Receiver 1 (GPRS)"),
            17: dict(message="Fail To Communicate IP Receiver 2 (GPRS)"),
            18: dict(message="IP Module No Service"),
            19: dict(message="IP Module Supervision Loss"),
            20: dict(message="Fail To Communicate IP Receiver 1 (IP)"),
            21: dict(message="Fail To Communicate IP Receiver 2 (IP)"),
            99: dict(message="Any new trouble event"),
        },
    ),
    45: dict(
        level=EventLevel.INFO,
        type="system",
        tags=["trouble", "restore"],
        sub={
            0: dict(message="Telephone line restore"),
            1: dict(message="AC failure restore"),
            2: dict(message="Battery failure restore"),
            3: dict(message="Auxiliary current overload restore"),
            4: dict(message="Bell current overload restore"),
            5: dict(message="Bell disconnected restore"),
            6: dict(message="Clock loss restore"),
            7: dict(message="Fire loop trouble restore"),
            8: dict(
                message="Fail to communicate to monitoring station telephone #1 restore"
            ),
            9: dict(
                message="Fail to communicate to monitoring station telephone #2 restore"
            ),
            11: dict(message="Fail to communicate to voice report restore"),
            12: dict(message="RF jamming restore"),
            13: dict(message="GSM RF jamming restore"),
            14: dict(message="GSM no service restore"),
            15: dict(message="GSM supervision lost restore"),
            16: dict(message="Fail To Communicate IP Receiver 1 (GPRS) restore"),
            17: dict(message="Fail To Communicate IP Receiver 2 (GPRS) restore"),
            18: dict(message="IP Module No Service restore"),
            19: dict(message="IP Module Supervision Loss restore"),
            20: dict(message="Fail To Communicate IP Receiver 1 (IP) restore"),
            21: dict(message="Fail To Communicate IP Receiver 2 (IP) restore"),
            99: dict(message="Any trouble event restore"),
        },
    ),
    46: dict(
        level=EventLevel.WARN,
        type="module",
        tags=["trouble"],
        sub={
            0: dict(message="Bus / EBus / Wireless module communication fault"),
            1: dict(message="Tamper trouble"),
            2: dict(message="Power fail"),
            3: dict(message="Battery failure"),
            99: dict(message="Any bus module new trouble event"),
        },
    ),
    47: dict(
        level=EventLevel.INFO,
        type="system",
        tags=["trouble", "restore"],
        sub={
            0: dict(message="Bus / EBus / Wireless module communication fault restore"),
            1: dict(message="Tamper trouble restore"),
            2: dict(message="Power fail restore"),
            3: dict(message="Battery failure restore"),
            99: dict(message="Any bus module trouble restored event"),
        },
    ),
    48: dict(
        level=EventLevel.INFO,
        type="system",
        sub={
            0: dict(message="System power up"),
            1: dict(message="Reporting test"),
            2: dict(message="Software log on"),
            3: dict(message="Software log off"),
            4: dict(message="Installer in programming mode"),
            5: dict(message="Installer exited programming mode"),
            6: dict(message="Maintenance in programming mode"),
            7: dict(message="Maintenance exited programming mode"),
            8: dict(message="Closing delinquency delay elapsed"),
            99: dict(message="Any special event"),
        },
    ),
    49: dict(
        level=EventLevel.WARN,
        changes=dict(zone_rf_low_battery_trouble=True),
        type="zone",
        tags=["trouble"],
        message="Low battery on zone {label}",
    ),
    50: dict(
        level=EventLevel.INFO,
        changes=dict(zone_rf_low_battery_trouble=False),
        type="zone",
        tags=["trouble", "restore"],
        message="Low battery on zone {label} restore",
    ),
    51: dict(
        level=EventLevel.WARN,
        changes=dict(zone_supervision_trouble=True),
        type="zone",
        tags=["trouble", "supervision"],
        message="Zone {label} supervision trouble",
    ),
    52: dict(
        level=EventLevel.INFO,
        changes=dict(zone_supervision_trouble=False),
        type="zone",
        tags=["trouble", "supervision", "restore"],
        message="Zone {label} supervision restore",
    ),
    53: dict(
        level=EventLevel.WARN,
        change=dict(module_supervision_trouble=True),
        type="module",
        tags=["trouble", "supervision"],
        message="Wireless module {label} supervision trouble",
        sub=event_opt1,
    ),
    54: dict(
        level=EventLevel.INFO,
        change=dict(module_supervision_trouble=False),
        type="module",
        tags=["trouble", "supervision", "restore"],
        message="Wireless module {label} supervision restore",
        sub=event_opt1,
    ),
    55: dict(
        level=EventLevel.WARN,
        change=dict(module_tamper_trouble=True),
        type="module",
        tags=["trouble", "tamper"],
        message="Wireless module {label} tamper trouble",
        sub=event_opt1,
    ),
    56: dict(
        level=EventLevel.INFO,
        change=dict(module_tamper_trouble=False),
        type="module",
        tags=["trouble", "tamper", "restore"],
        message="Wireless module {label} tamper restore",
    ),
    57: dict(
        level=EventLevel.CRITICAL,
        type="user",
        tags=["alarm", "medic", "trigger"],
        message="Non-medical alarm (paramedic) by {label}",
    ),
    58: dict(level=EventLevel.DEBUG, type="zone", message="Zone {label} forced"),
    59: dict(level=EventLevel.DEBUG, type="zone", message="Zone {label} included"),
    64: dict(
        level=EventLevel.DEBUG,
        type="system",
        sub={
            0: dict(message="Follow Arm LED status"),
            1: dict(message="PGM pulse fast in alarm"),
            2: dict(message="PGM pulse fast in exit delay below 10 sec."),
            3: dict(message="PGM pulse slow in exit delay over 10 sec."),
            4: dict(message="PGM steady ON if armed"),
            5: dict(message="PGM OFF if disarmed"),
        },
    ),
}

_labels = {
    "zone": "Zone Number {label}",
    "pgm": "PGM Number {label}",
    "user": "User {label}",
    "keypad": "Wireless Keypad {label}",
    "repeater": "Wireless Repeater {label}",
    "siren": "Wireless Siren {label}",
}
